@@ -1,17 +1,17 @@
-/************************************************************
+/**********************************************************************
  *
- minires 0.96         stub synchronous resolver for Cygwin
+ minires 1.00         stub synchronous resolver for Cygwin
  Pierre A. Humblet
- # June 2003
+ October 2004
 
- Copyright (c) 2001, 2002, 2003 Pierre A. Humblet
+ Copyright (c) 2001, 2002, 2003, 2004 Pierre A. Humblet
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version
 
-************************************************************/
-#include "resolver.h"
+**********************************************************************/
+#include "minires.h"
 
 /***********************************************************************
 
@@ -22,7 +22,7 @@ Utilities
 
 dprintf
 ***********************************************************************/
-void dprintf(char * format, ...)
+void minires_dprintf(char * format, ...)
 {
   va_list args;
   
@@ -35,8 +35,8 @@ void dprintf(char * format, ...)
 /***********************************************************************
 
 scanline
-Put pointers in list[] to the beginning of each space delimited word in "in",
-and put the lengths in sizes[]
+Put pointers in list[] to the beginning of each space or comma delimited
+word in "in", and put the lengths in sizes[] (counting the final 0).
 Return the number of words found  
 ***********************************************************************/
 static int scanline(char * in, char **list, int * sizes, int maxnum)
@@ -44,132 +44,168 @@ static int scanline(char * in, char **list, int * sizes, int maxnum)
   int i;
   char * startp;
   for (i = 0; i < maxnum; i++) {
-    while((*in) && isspace(*in)) in++;
+    while((*in) && (isspace(*in) || *in == ',')) in++;
     if (*in == 0) 
       break;
     startp = in++;
-    while((*in) && !isspace(*in)) in++;
+    while((*in) && !isspace(*in) && *in != ',') in++;
     list[i] = startp;
-    sizes[i] = in - startp; 
+    sizes[i] = in - startp + 1; 
+    if (*in)
+      *in++ = 0;
   }
   return i;
 }
 
 /***********************************************************************
 
-Make string
+Read the search string.
 
 ***********************************************************************/
-char * make_string( char * from, int len, int debug)
+void minires_get_search(char * string, res_state statp)
 {
-  char * ptr;
-  if ( len < 0 ) 
-    len = strlen(from); 
-  if ((ptr = (char *) malloc(len+1)) == 0) {
-    DPRINTF(debug, "malloc: %s\n", strerror(errno));
+  char * words[MAXDNSRCH+1], * ptr;
+  int sizes[MAXDNSRCH+1];
+  int i, j, debug = statp->options & RES_DEBUG;
+
+  i = scanline(string, words, sizes, MAXDNSRCH+1);
+  ptr = statp->defdname;
+  for (j = 0; j < i; j++) {
+    if (j < MAXDNSRCH 
+	&& ptr + sizes[j] < &statp->defdname[DIM(statp->defdname)]) {
+      statp->dnsrch[j] = strcpy(ptr, words[j]);
+      statp->dnsrch[j+1] = NULL;
+      ptr += sizes[j];
+      DPRINTF(debug, "search \"%s\"\n", words[j]);
+    }
+    else
+      DPRINTF(debug, "no space for \"%s\"\n", words[j]);
   }
-  else {
-    memcpy(ptr, from, len);
-    ptr[len] = 0;
-  }
-  return ptr;
-}   
+}
 
 /***********************************************************************
 
-Read the LOCALDOMAIN string (search list) 
+Read options
+
 
 ***********************************************************************/
-static void get_env_domain( char * envstring, char ** dnsrch, int debug)
+static void get_options(res_state statp, int i, char **words)
 {
-  char * words[MAXDNSRCH];
-  int sizes[MAXDNSRCH];
-  int i, j;
-  i = scanline(envstring, words, sizes, MAXDNSRCH);
-  for (j = 0; j < MAXDNSRCH; j++) {
-    if (dnsrch[j] != NULL ) {
-      DPRINTF(debug, "removing \"%s\"\n", dnsrch[j]);
-      free(dnsrch[j]);  /* was already allocated */
+  char *ptr;
+  int value;
+
+  while (i-- > 0) {
+    if (!strcasecmp("debug", words[i])) {
+      statp->options |= RES_DEBUG;
+      DPRINTF(statp->options & RES_DEBUG, "%s: 1\n", words[i]);
+      continue;
     }
-    if (j < i) {
-      dnsrch[j] = make_string(words[j], sizes[j], debug);
-      DPRINTF(debug, "LOCALDOMAIN \"%s\"\n", dnsrch[j]);
+    if (!strcasecmp("osquery", words[i])) {
+      statp->use_os = 1;
+      DPRINTF(statp->options & RES_DEBUG, "%s: 1\n", words[i]);
+      continue;
     }
-    else dnsrch[j] = 0;
+
+    if ((ptr = strchr(words[i], ':'))) {
+      *ptr++ = 0;
+      value = atoi(ptr);
+      /* Not supported 
+	 if (!strcasecmp("ndots", words[i])) {
+	 statp->ndots = value;
+	 continue;
+	 }
+      */
+      if (!strcasecmp("retry", words[i])) {
+	if (value < 1)
+	  value = 1;
+	statp->retry = value;
+	DPRINTF(statp->options & RES_DEBUG, "%s: %d\n", words[i], value);
+	continue;
+      }
+      if (!strcasecmp("retrans", words[i])) {
+	if (value < 1)
+	  value = 1;
+	statp->retrans = value;
+	DPRINTF(statp->options & RES_DEBUG, "%s: %d\n", words[i], value);
+	continue;
+      }
+    }
+    DPRINTF(statp->options & RES_DEBUG, "unknown option: \"%s\"\n", words[i]);
   }
 }
 
 /***********************************************************************
 
 Read the resolv.conf file.
-We only look for nameserver, domain and search
+We only look for nameserver, domain, search and options
 
 ***********************************************************************/
-#if MAXNS > MAXDNSRCH
+#if MAXNS > MAXDNSRCH + 1
 #define MAXSIZE MAXNS
 #else 
-#define MAXSIZE MAXDNSRCH
+#define MAXSIZE MAXDNSRCH + 1 /* Make unused one visible */
 #endif
-static int get_resolv(struct sockaddr_in * SockAddrPtr, char ** dnsrch, int debug)
+static void get_resolv(res_state statp)
 {
   FILE * fd;
-  char *words[MAXSIZE + 1], line[4096];
-  int sizes[MAXSIZE + 1];
-  int i, j, ns = 0;
-  if ((fd = fopen(_PATH_RESCONF, "r")) == NULL) 
-    return 0;
+  char *words[MAXSIZE + 1], line[4096], *ptr;
+  int sizes[DIM(words)];
+  int i, j, ns = 0, have_search, have_address, debug = statp->options & RES_DEBUG;
+
+  fd = fopen(_PATH_RESCONF, "r");
+  DPRINTF(debug, _PATH_RESCONF ": %s\n", fd?"OK":strerror(errno));
+  if (fd == NULL)
+    return;
+
+  statp->use_os = 0; /* Do not use os_query, except if allowed by "options" */
+  have_search = (statp->dnsrch[0] != NULL);
+  have_address = (statp->nscount != 0);
+
   while ( fgets(line, sizeof(line), fd) != 0) {
     DPRINTF(debug, "resolv.conf %s", line);
-    if ((i = scanline(line, words, sizes, MAXSIZE + 1)) > 0) {
-      if (strncasecmp("nameserver", words[0], sizes[0]) == 0) {
-        for ( j = 1; j < i ; j++) {
+    if ((i = scanline(line, words, sizes, DIM(words))) > 0) {
+      if (!have_address 
+	  && !strncasecmp("nameserver", words[0], sizes[0])) {
+	for ( j = 1; j < i ; j++) {
 	  unsigned int address;
-          *(words[j] + sizes[j]) = 0;
 	  address = inet_addr(words[j]);
-          if (address == -1) {
+	  if (address == -1) {
 	    DPRINTF(debug, "invalid server \"%s\"\n", words[j]);
 	  }
 	  else if (ns >= MAXNS) {
 	    DPRINTF(debug, "no space for server \"%s\"\n", words[j]);
 	  }
 	  else {
-	    SockAddrPtr[ns++].sin_addr.s_addr = address;
+	    statp->nsaddr_list[ns++].sin_addr.s_addr = address;
+	    statp->nscount++;
 	    DPRINTF(debug, "server \"%s\"\n", words[j]);
 	  }
-        }
+	}
       }
-      else if (strncasecmp("search", words[0], sizes[0]) == 0) {
-        for (j = 0; j < MAXDNSRCH; j++) {
-          if (dnsrch[j] != NULL ) {
-	    DPRINTF(debug, "removing \"%s\"\n", dnsrch[j]);
-	    free(dnsrch[j]);  /* was already allocated */
+      else if (!have_search 
+	       && (!strncasecmp("search", words[0], sizes[0])
+		   || !strncasecmp("domain", words[0], sizes[0]))) {
+	ptr = statp->defdname;
+	for (j = 0; j + 1 < i; j++) {
+	  if (j < MAXDNSRCH
+	      && ptr + sizes[j + 1] < &statp->defdname[DIM(statp->defdname)]) {
+	    statp->dnsrch[j] = strcpy(ptr, words[j+1]);
+	    statp->dnsrch[j+1] = 0;
+	    ptr += sizes[j+1];
+	    DPRINTF(debug, "domain|search \"%s\"\n", statp->dnsrch[j]);
 	  }
-          if (j < i -1) { 
-	    dnsrch[j] = make_string(words[j+1], sizes[j+1], debug);
-	    DPRINTF(debug, "domain \"%s\"\n", dnsrch[j]);
+	  else {
+	    DPRINTF(debug, "no space for \"%s\"\n", words[j+1]);
 	  }
-          else 
-	    dnsrch[j] = 0;
-        }
-      }
-      /* Domain goes into 0th element of dnssrch */
-      else if (strncasecmp("domain", words[0], sizes[0]) == 0) {
-        for (j = 0; j < MAXDNSRCH; j++) {
-          if (dnsrch[j] != NULL ) {
-	    DPRINTF(debug, "removing \"%s\"\n", dnsrch[j]);
-	    free(dnsrch[j]);  /* was already allocated */
-	  }
-          if (j < 1) {
-	    dnsrch[j] = make_string(words[1], sizes[1], debug);
-	    DPRINTF(debug, "domain \"%s\"\n", dnsrch[j]);
-	  }
-          else dnsrch[j] = 0;
-        }
+	}
       }
+      /* Options line */
+      else if (!strncasecmp("options", words[0], sizes[0]))
+	get_options(statp, i - 1, &words[1]);
     }
   }
   fclose(fd);
-  return ns;
+  return;
 }
 
 /****************************************************************************/
@@ -206,7 +242,7 @@ static int open_sock(struct sockaddr_in *CliAddr, int debug)
  Undocumented but public. Accessed through _res
 
  *****************************************************************/
-static struct __res_state res = {};
+static struct __res_state res;
 struct __res_state *__res_state(void)
 {
   return & res;
@@ -219,37 +255,42 @@ struct __res_state *__res_state(void)
  *****************************************************************/
 int res_ninit(res_state statp)
 {
-  int i, num, num2;
+  int i;
   char * ptr;
   
   statp->res_h_errno = NETDB_SUCCESS;
-  /* Get the domainname (dnsrch[0]) 
-     the search list
-     the number of nameservers and their IP addresses (Network Byte Order) */
-  for (i = 0; i < MAXDNSRCH + 1; i++)  statp->dnsrch[i] = 0;
-  /* Get domain and dns server info from an os-specific routine */
-  num = get_dns_info(statp->dnsrch, statp->nsaddr_list, MAXNS, statp->options & RES_DEBUG );
-  /* resolv.conf (dns, domain & srch)*/
-  num2 = get_resolv(statp->nsaddr_list, statp->dnsrch, statp->options & RES_DEBUG); 
-  if (num2 > 0) 
-    num = num2; /* Give preference to resolv.conf */
-  if ((ptr = getenv("LOCALDOMAIN")) != 0 ) 
-    get_env_domain( ptr, statp->dnsrch, statp->options & RES_DEBUG); /* domain & srch */
-  if (num <= 0) {
+  statp->nscount = 0;
+  statp->os_query = NULL;
+  statp->retrans = RES_TIMEOUT; /* timeout in seconds */
+  statp->retry = RES_MAXRETRY;  /* max number of retries */
+  statp->use_os = 1;            /* use os_query if available and allowed by get_resolv */
+  statp->mypid = -1;
+  statp->sockfd = -1;
+
+  for (i = 0; i < DIM(statp->dnsrch); i++)  statp->dnsrch[i] = 0;
+
+  /* Get search list from LOCALDOMAIN */
+  if ((ptr = getenv("LOCALDOMAIN")) != 0 ) {
+    DPRINTF(statp->options & RES_DEBUG, "LOCALDOMAIN \"%s\"\n", ptr);
+    minires_get_search(ptr, statp); /* domain or dnsrch */
+  }
+  /* resolv.conf (dns servers & search list)*/
+  get_resolv(statp);
+  /* Get dns servers and search list from an os-specific routine, set os_query */
+  get_dns_info(statp);
+      
+  if (statp->nscount == 0 && !statp->os_query) {
+    errno = ENONET;
     statp->res_h_errno = NETDB_INTERNAL;
     DPRINTF(statp->options & RES_DEBUG, "no dns server found\n"); 
     return -1;
   }
-  for (i = 0; i < num; i++) {
+  for (i = 0; i < statp->nscount; i++) {
     statp->nsaddr_list[i].sin_family = AF_INET;
     statp->nsaddr_list[i].sin_port = htons(NAMESERVER_PORT);
     bzero(statp->nsaddr_list[i].sin_zero, sizeof(statp->nsaddr_list[i].sin_zero));
   }
-  statp->sockfd = -1;
-  statp->nscount = num;         /* number of servers */
-  statp->retrans = RES_TIMEOUT; /* timeout in seconds */
-  statp->retry = RES_MAXRETRY;  /* max number of retries */
-  /* only debug may be set before calling init */
+  /* Only debug may be set before calling init */
   statp->options &= RES_DEBUG;
   statp->options |= RES_INIT | RES_DEFAULT;
   return 0;
@@ -283,6 +324,75 @@ void res_close()
   res_nclose(& res);
 }  
 
+/*****************************************************************
+ *
+ get_tcp_buf()
+
+ *****************************************************************/
+static int get_tcp_buf(int fd, unsigned char *buf, int size, int debug)
+{
+  int res;
+  while (size > 0) {
+    if ((res = read(fd, buf, size)) < 0) {
+      DPRINTF(debug, "read: %s\n", strerror(errno));
+      return -1;
+    }
+    DPRINTF(debug, "read %d out of %d\n", res, size);
+    size -= res;
+    buf += res;
+  }
+  return 0;
+}
+
+/*****************************************************************
+ *
+ get_tcp()
+
+ *****************************************************************/
+static int get_tcp(struct sockaddr_in *CliAddr, 
+		   const unsigned char * MsgPtr, int MsgLength,
+		   unsigned char * AnsPtr, int AnsLength, int debug)
+{
+  int fd, res = -1;
+  unsigned short ans_length;
+  union {short len; u_char buf[sizeof(short)];} len_buf;
+
+  DPRINTF(debug, "retrying with TCP\n");
+
+  /* Create a tcp socket */
+  if ((fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
+    DPRINTF(debug, "socket: %s\n", strerror(errno));
+    return -1;
+  }
+
+  if (connect(fd, (struct sockaddr *)  CliAddr, sizeof(* CliAddr)) < 0) {
+    DPRINTF(debug, "connect: %s\n", strerror(errno));
+    goto done;
+  }
+
+  /* Send the length then the message */
+  len_buf.len = htons(MsgLength);
+  if (write(fd, len_buf.buf, sizeof(len_buf)) != sizeof(len_buf)
+      || write(fd, MsgPtr, MsgLength) != MsgLength) {
+    DPRINTF(debug, "write: %s\n", strerror(errno));
+    goto done;
+  }
+
+  /* Read the answer length */
+  if (get_tcp_buf(fd, len_buf.buf, sizeof(len_buf), debug))
+    goto done;
+  ans_length = ntohs(len_buf.len);
+
+  /* Read the answer */
+  if (get_tcp_buf(fd, AnsPtr, MIN(ans_length, AnsLength), debug)) 
+    goto done;
+  res = ans_length;
+
+ done:
+  close (fd);
+  return res;
+}
+
 /*****************************************************************
  **
  res_send
@@ -295,10 +405,11 @@ int res_nsend( res_state statp, const unsigned char * MsgPtr,
 {
   /* Current server, shared by all tasks */
   volatile static unsigned int SServ = 0XFFFFFFFF;
+  int tcp;
 
   fd_set fdset_read; 
   int rslt, addrLen, transNum, wServ;
-  struct sockaddr mySockAddr, dnsSockAddr;
+  struct sockaddr_in mySockAddr, dnsSockAddr;
   struct timeval timeOut;
   
   statp->res_h_errno = NETDB_SUCCESS;
@@ -306,7 +417,7 @@ int res_nsend( res_state statp, const unsigned char * MsgPtr,
     return -1;
 
   /* Close the socket if it had been opened before a fork.
-     Reuse of pid's doesn't hurt */
+     Reuse of pid's cannot hurt */
   if ((statp->sockfd != -1) && (statp->mypid != getpid())) {
     res_nclose(statp);
   }
@@ -314,7 +425,7 @@ int res_nsend( res_state statp, const unsigned char * MsgPtr,
   /* Open a socket for this process */
   if (statp->sockfd == -1) {
     /* Create a socket and bind it (to any port) */
-    statp->sockfd = open_sock((struct sockaddr_in *) & mySockAddr, 
+    statp->sockfd = open_sock(& mySockAddr, 
 			      statp->options & RES_INIT);
     DPRINTF(statp->options & RES_DEBUG, "open sockfd %d: %s\n", 
 	    statp->sockfd, (statp->sockfd > 0)?"OK":strerror(errno));
@@ -322,16 +433,16 @@ int res_nsend( res_state statp, const unsigned char * MsgPtr,
       statp->res_h_errno = NETDB_INTERNAL;
       return -1;
     }
-    /* Set close on exec bit */
+    /* Set close on exec flag */
     if (fcntl(statp->sockfd, F_SETFD, 1) == -1) {
       DPRINTF(statp->options & RES_DEBUG, "fcntl: %s\n", 
-                strerror(errno));
+	      strerror(errno));
       statp->res_h_errno = NETDB_INTERNAL;
       return -1;
     }
     statp->mypid = getpid();
     if (SServ == 0XFFFFFFFF) /* Pseudo random */
-      SServ =  statp->mypid % statp->retry;
+      SServ =  statp->mypid % statp->nscount;
   }
 
   transNum = 0;  
@@ -342,7 +453,7 @@ int res_nsend( res_state statp, const unsigned char * MsgPtr,
     /* Send the message */
     rslt = sendto(statp->sockfd, MsgPtr, MsgLength, 0, 
 		  (struct sockaddr *) &statp->nsaddr_list[wServ],
-		  sizeof(struct sockaddr));
+		  sizeof(struct sockaddr_in));
     DPRINTF(statp->options & RES_DEBUG, "sendto: server %08x sockfd %d %s\n",
 	    statp->nsaddr_list[wServ].sin_addr.s_addr, 
 	    statp->sockfd, (rslt == MsgLength)?"OK":strerror(errno));
@@ -361,63 +472,83 @@ int res_nsend( res_state statp, const unsigned char * MsgPtr,
     if ( rslt == 0 ) { /* Timeout */
       DPRINTF(statp->options & RES_DEBUG, "timeout for server %08x\n",
 	      statp->nsaddr_list[wServ].sin_addr.s_addr);
+      continue;
     }
     else if ((rslt != 1) || (FD_ISSET(statp->sockfd, &fdset_read) == 0)) {
       DPRINTF(statp->options & RES_DEBUG, "select: %s\n", strerror(errno));
       statp->res_h_errno = NETDB_INTERNAL;
       return -1;
     }
-    else {
-      addrLen = sizeof(struct sockaddr);
-      rslt = recvfrom( statp->sockfd, AnsPtr, AnsLength, 0, 
-		       & dnsSockAddr, & addrLen);
-      if (rslt <= 0) {
-        DPRINTF(statp->options & RES_DEBUG, "recvfrom: %s\n", 
-		strerror(errno));
-        statp->res_h_errno = NETDB_INTERNAL;
-        return -1;
-      }
+
+    addrLen = sizeof(dnsSockAddr);
+    rslt = recvfrom( statp->sockfd, AnsPtr, AnsLength, 0, 
+		     (struct sockaddr *) & dnsSockAddr, & addrLen);
+    if (rslt <= 0) {
+      DPRINTF(statp->options & RES_DEBUG, "recvfrom: %s\n", 
+	      strerror(errno));
+      statp->res_h_errno = NETDB_INTERNAL;
+      return -1;
+    }
+    /* 
+       Prepare to retry with tcp 
+    */
+    for (tcp = 0; tcp < 2; tcp++) {
       /* Check if this is the message we expected */
-// #define NETDB_INTERNAL -1 /* see errno */
-// #define NETDB_SUCCESS   0 /* no problem */
-// #define HOST_NOT_FOUND  1 /* Authoritative Answer Host not found */
-// #define TRY_AGAIN       2 /* Non-Authoritive Host not found, or SERVERFAIL */
-// #define NO_RECOVERY     3 /* Non recoverable errors, FORMERR, REFUSED, NOTIMP */
-// #define NO_DATA         4 /* Valid name, no data record of requested type */
       if ((*MsgPtr == *AnsPtr)     /* Ids match */
 	  && (*(MsgPtr + 1) == *(AnsPtr + 1))
-	  && (rslt >= MsgLength )     /* At least as long */
-	  && (memcmp(MsgPtr + 12, AnsPtr + 12, MsgLength - 12) == 0)
+/* We have stopped checking this because the question may not be present on error,
+   in particular when the name in the question is not a valid name.
+   Simply check that the header is present. */
+	  && (rslt >= HFIXEDSZ)
+/*        && (rslt >= MsgLength )
+	  && (memcmp(MsgPtr + HFIXEDSZ, AnsPtr + HFIXEDSZ, MsgLength - HFIXEDSZ) == 0) */
 	  && ((AnsPtr[2] & QR) != 0)) {
-        DPRINTF(statp->options & RES_DEBUG, 
-		"answer from %08x to query to %08x. Error %d. Count %d\n",
-		((struct sockaddr_in *)&dnsSockAddr)->sin_addr.s_addr, 
-		statp->nsaddr_list[wServ].sin_addr.s_addr,
+
+	DPRINTF(statp->options & RES_DEBUG, 
+		"answer %u from %08x. Error %d. Count %d.\n",
+		rslt, dnsSockAddr.sin_addr.s_addr, 
 		AnsPtr[3] & ERR_MASK, AnsPtr[6]*256 + AnsPtr[7]);
-        if ((AnsPtr[3] & ERR_MASK) == NOERROR) {
-          if ((AnsPtr[6]|AnsPtr[7])!= 0) 
+#if 0
+ NETDB_INTERNAL -1 /* see errno */
+ NETDB_SUCCESS   0 /* no problem */
+ HOST_NOT_FOUND  1 /* Authoritative Answer Host not found */
+ TRY_AGAIN       2 /* Non-Authoritive Host not found, or SERVERFAIL */
+                         Also seen returned by some servers when the name is too long
+ NO_RECOVERY     3 /* Non recoverable errors, FORMERR, REFUSED, NOTIMP */
+ NO_DATA         4 /* Valid name, no data record of requested type */
+#endif
+	if ((AnsPtr[3] & ERR_MASK) == NOERROR) {
+	  if ((AnsPtr[2] & TC) && !(statp->options & RES_IGNTC)) { /* Truncated. Try TCP */
+	    rslt = get_tcp(&statp->nsaddr_list[wServ], MsgPtr, MsgLength,
+			   AnsPtr, AnsLength, statp->options & RES_DEBUG);
+	    continue; 
+	  }
+	  else if ((AnsPtr[6] | AnsPtr[7])!= 0) 
 	    return rslt;
-          else 
+	  else 
 	    statp->res_h_errno = NO_DATA;
-        }
-        else {
-          /* return HOST_NOT_FOUND even for non-authoritative answers */
-          if ((AnsPtr[3] & ERR_MASK) == NXDOMAIN) 
+	}
+	else {
+	  /* return HOST_NOT_FOUND even for non-authoritative answers */
+	  if ((AnsPtr[3] & ERR_MASK) == NXDOMAIN) 
 	    statp->res_h_errno = HOST_NOT_FOUND;
-          else if ((AnsPtr[3] & ERR_MASK) == SERVFAIL) 
+	  else if ((AnsPtr[3] & ERR_MASK) == SERVFAIL) 
 	    statp->res_h_errno = TRY_AGAIN;
-          else statp->res_h_errno = NO_RECOVERY;
-        }
-        return -1;
+	  else 
+	    statp->res_h_errno = NO_RECOVERY;
+	}
+	return -1;
       }
-      else 
-	DPRINTF(statp->options & RES_DEBUG, "unexpected answer from %x to query to %x\n",
-		((struct sockaddr_in *)&dnsSockAddr)->sin_addr.s_addr, 
+      else {
+	DPRINTF(statp->options & RES_DEBUG, "unexpected answer %u from %x to query to %x\n",
+		rslt, dnsSockAddr.sin_addr.s_addr, 
 		statp->nsaddr_list[wServ].sin_addr.s_addr);
-    }
+	break;
+      }
+    } /* TCP */
   }
-  DPRINTF(statp->options & RES_DEBUG, "too many timeouts\n");
-  statp->res_h_errno = TRY_AGAIN; /* Too many timeouts */
+  DPRINTF(statp->options & RES_DEBUG, "too many retries\n");
+  statp->res_h_errno = TRY_AGAIN;
   return -1;
 }
 
@@ -435,64 +566,37 @@ int res_send( const unsigned char * MsgPtr, int MsgLength,
 
  Return: packet size
         -1 name format is incorrect 
-        (enforces rule on max label and domain lengths) 
 *****************************************************************/
 int res_nmkquery (res_state statp,
 		  int op, const char * dnameptr, int qclass, int qtype, 
 		  const unsigned char * dataptr, int datalen,
 		  const unsigned char * newrr, unsigned char * buf, int buflen)
 {
-  unsigned char * cptr, * lptr;
-  int len;
+  int i, len;
   short id;
 
   if (op == QUERY) {
-    /* Check if name is too long
-       nameserver.h defines MAXDNAME and MAXCDNAME, with different systems
-       having different MAXDNAME. MAXCDNAME is 255, RFC 1035, p. 9
-       A '.' is added at the end of dnameptr if none is present
-       The size of the encoded name is 2 char longer than the original,
-       if it doesn't have a final .  */
-    len = strlen(dnameptr);
-    if ((len >= (MAXCDNAME-1))  /* <= MAXDCNAME - 2 is OK */ 
-        && ((len > (MAXCDNAME-1)) || dnameptr[MAXCDNAME] == 0  )) {
-      DPRINTF(statp->options & RES_DEBUG, "total length too large\n");
-      statp->res_h_errno = NO_RECOVERY;
+    /* Write the name and verify buffer length */
+    len = dn_comp(dnameptr, buf + HFIXEDSZ, buflen - HFIXEDSZ - QFIXEDSZ, NULL, NULL);
+    if (len < 0) {
+      DPRINTF(statp->options & RES_DEBUG, 
+	      "\"%s\" invalid or buffer too short\n", dnameptr);
+      statp->res_h_errno = NETDB_INTERNAL;
       return -1;
     }
-    cptr = buf;
     /* Fill the header */
-    id = statp->id++;
-    PUTSHORT(id, cptr);
-    PUTSHORT(RD, cptr); 
-    PUTSHORT(1, cptr); /* Number of */
-    PUTSHORT(0, cptr); /* Number of */
-    PUTSHORT(0, cptr); /* Number of */
-    PUTSHORT(0, cptr); /* Number of */
-    /* Put the name */
-    do {
-      /* New label */
-      lptr = cptr++;   /* Save the length ptr */
-      len = MAXLABEL;
-      while ((*dnameptr != '.') && (*dnameptr != 0)) {
-        if (len-- == 0) {
-          DPRINTF(statp->options & RES_DEBUG, "label too long\n");
-          statp->res_h_errno = NO_RECOVERY;
-          return -1;
-        }
-        *cptr++ = *dnameptr++;
-      }
-      *lptr = MAXLABEL - len; /* Write label length */
-      if (*lptr == 0) 
-	goto fill_done; /* Have 0 */
-    } while(*dnameptr++ != 0); /* Skip over . */
-    *cptr++ = 0;               /* Final 0 */
-  fill_done:
-  
+    id = statp->id;
+    PUTSHORT(id, buf);
+    PUTSHORT(RD, buf); 
+    PUTSHORT(1, buf); /* Number of questions */
+    for (i = 0; i < 3; i++)
+      PUTSHORT(0, buf); /* Number of answers */
+
     /* Write qtype and qclass */
-    PUTSHORT(qtype, cptr);
-    PUTSHORT(qclass, cptr);
-    return cptr - buf; /* packet size */
+    buf += len;
+    PUTSHORT(qtype, buf);
+    PUTSHORT(qclass, buf);
+    return len + 16; /* packet size */
   }
   else { /* Not implemented */
     errno = ENOSYS;
@@ -506,7 +610,7 @@ int res_mkquery (int op, const char * dnameptr, int qclass, int qtype,
 		 const unsigned char * newrr, unsigned char * buf, int buflen)
 {
   int r = res_nmkquery (& res, op, dnameptr, qclass, qtype, 
-		       dataptr, datalen, newrr, buf, buflen);
+			dataptr, datalen, newrr, buf, buflen);
   h_errno = res.res_h_errno;
   return r;
 
@@ -521,11 +625,16 @@ int res_mkquery (int op, const char * dnameptr, int qclass, int qtype,
 int res_nquery( res_state statp, const char * DomName, int Class, int Type, 
 		unsigned char * AnsPtr, int AnsLength)
 {
-  char packet[PACKETSZ];
+  u_char packet[PACKETSZ];
   int len;
   
-  DPRINTF(statp->options & RES_DEBUG, "query %s type %d\n", DomName, Type);
-  statp->res_h_errno = NETDB_SUCCESS;  
+  DPRINTF(statp->options & RES_DEBUG, "query \"%s\" type %d\n", DomName, Type);
+  statp->res_h_errno = NETDB_SUCCESS;
+
+  /* If a hook exists to a native implementation, use it */
+  if (statp->os_query)
+    return ((os_query_t *) statp->os_query)(statp, DomName, Class, Type, AnsPtr, AnsLength);
+
   if ((len = res_nmkquery (statp, QUERY, DomName, Class, Type, 
 			   0, 0, 0, packet, PACKETSZ)) < 0)
     return -1;
@@ -547,26 +656,31 @@ int res_query( const char * DomName, int Class, int Type, unsigned char * AnsPtr
 int res_nquerydomain( res_state statp, const char * Name, const char * DomName,
 		      int Class, int Type, unsigned char * AnsPtr, int AnsLength)
 {
-  char fqdn[MAXCDNAME+1], *ptr;
+  char fqdn[MAXDNAME], *ptr;
   int nlen;
 
   if (!DomName)
     ptr = (char *) Name;
-  else if ((nlen = strlen(Name)) >= sizeof(fqdn) - 1) {
-    DPRINTF(statp->options & RES_DEBUG, "querydomain: name too long\n");
-    statp->res_h_errno = NO_RECOVERY;
-    return -1;
-  }
+  else if ((nlen = strlen(Name)) >= sizeof(fqdn) - 1)
+    goto error;
   else {
     strcpy(fqdn, Name);
     ptr = &fqdn[nlen];
     if (nlen && *(ptr - 1) != '.') 
       *(ptr++ - 1) = '.';
-    /* Length checking and res_h_errno setting are done in res_mkquery */
-    strlcpy(ptr, DomName, sizeof(fqdn) - (ptr - fqdn));
+    fqdn[sizeof(fqdn) - 1] = 0;
+    strncpy(ptr, DomName, sizeof(fqdn) - (ptr - fqdn));
+    if (fqdn[sizeof(fqdn) - 1])
+      goto error;
     ptr = fqdn;
   }
   return res_nquery(statp, ptr, Class, Type, AnsPtr, AnsLength);
+
+ error:
+  DPRINTF(statp->options & RES_DEBUG, "querydomain: name too long\n");
+  errno = EINVAL;
+  statp->res_h_errno = NETDB_INTERNAL;;
+  return -1;
 }
 
 int res_querydomain( const char * Name, const char * DomName, int Class,
@@ -588,36 +702,48 @@ int res_nsearch( res_state statp, const char * DomName, int Class, int Type,
 		 unsigned char * AnsPtr, int AnsLength)
 {
   int len, stat, i;
-  char fullDomName[MAXCDNAME + 2], *ptr; /* Space for 0 + space for overflow detection */
+  char fullDomName[MAXDNAME], *ptr, *sptr;
+
+  DPRINTF(statp->options & RES_DEBUG, "search \"%s\" type %d\n", DomName, Type);
 
-  DPRINTF(statp->options & RES_DEBUG, "search %s type %d\n", DomName, Type);
   if (((statp->options & RES_INIT) == 0) && (res_ninit(statp) != 0)) 
     return -1;
-  /* Straight query. Will set res_h_errno */
-  stat = res_nquery( statp, DomName, Class, Type, AnsPtr, AnsLength);
-  /* Check if will skip search */
-  if ((statp->res_h_errno != HOST_NOT_FOUND)         /* Success, or bad error */
-      || (((statp->options & RES_DNSRCH) == 0)       /* No search */
-	  && (((statp->options & RES_DEFNAMES) == 0) /* and no def domain */
-	      || (strchr(DomName, '.') != NULL)))    /*   with no dot */
-      || (statp->dnsrch[0] == NULL))         /* No search list */
-    return stat;  
 
-  /* Must search */
-  len = strlen(DomName); /* HOST_NOT_FOUND implies len <= MAXLABEL */
+  stat = res_nquery( statp, DomName, Class, Type, AnsPtr, AnsLength);
+
+  /* Check if will skip search */
+  if (statp->res_h_errno != HOST_NOT_FOUND               /* Success or hard failure */
+      || ((ptr = strrchr(DomName, '.')) && (!*(ptr+1)))  /* Final dot */ 
+      || (((statp->options & RES_DNSRCH) == 0)           /* Or no search */
+	  && ((ptr != NULL)                              /*  And some dot */
+	      || ((statp->options & RES_DEFNAMES) == 0)))/*    or no def domain */
+      || (!(sptr = statp->dnsrch[0])))                 
+    return stat;
+
+  len = strlen(DomName);
+  if (len >= MAXDNAME - 1) /* Space for next dot */
+    goto error;
   strcpy(fullDomName, DomName); 
   fullDomName[len++] = '.';
-  fullDomName[MAXCDNAME+1] = 0; /* Overflow caught by res_mkquery() */
+  fullDomName[MAXDNAME - 1] = 0; /* Overflow indicator */
   i = 0;
-  do { /* Know  statp->dnsrch[i] != NULL */
-    ptr = statp->dnsrch[i];
-    strncpy(fullDomName + len, ptr, (sizeof(fullDomName)-1) - len);
-    stat = res_nquery( statp, fullDomName, Class, Type, AnsPtr, AnsLength);
-  } while ((statp->dnsrch[++i] != NULL)
-           && (statp->res_h_errno == HOST_NOT_FOUND)    /* Severe failure or host found */
-           && ((statp->options & RES_DNSRCH) != 0));
+  do {
+    strncpy(fullDomName + len, sptr, MAXDNAME - len);
+    if (fullDomName[MAXDNAME - 1])
+      goto error;      
+    stat = res_nquery(statp, fullDomName, Class, Type, AnsPtr, AnsLength);
+  } while ((sptr = statp->dnsrch[++i]) != NULL
+	   && statp->res_h_errno == HOST_NOT_FOUND
+	   && (statp->options & RES_DNSRCH) != 0);
+
   /* Return last stat */
   return stat;
+
+ error:
+  DPRINTF(statp->options & RES_DEBUG, "name too long during search\n");
+  errno = EINVAL;
+  statp->res_h_errno = NETDB_INTERNAL;
+  return -1;
 }
 
 int res_search( const char * DomName, int Class, int Type, 
@@ -637,10 +763,13 @@ int res_search( const char * DomName, int Class, int Type,
 int dn_expand(const unsigned char *msg, const unsigned char *eomorig,
               const unsigned char *comp_dn, char *exp_dn, int length)
 {
-  int len, complen = 0;
+  unsigned int len, complen = 0;
   const unsigned char *comp_dn_orig = comp_dn;
-//  char * exp_start = exp_dn;
+/*  char * exp_start = exp_dn; */
 
+  errno = EINVAL;
+  if (comp_dn >= eomorig)
+    goto expand_fail;
   if ((len = *comp_dn++) == 0)       /* Weird case */
     exp_dn++;
   else do {
@@ -667,11 +796,11 @@ int dn_expand(const unsigned char *msg, const unsigned char *eomorig,
   *(--exp_dn) = 0;
   if (!complen) 
     complen = comp_dn - comp_dn_orig;
-//  fprintf(stderr, "dn_expand %s\n", exp_start);
+/*  fprintf(stderr, "dn_expand %s\n", exp_start); */
   return complen;
   
 expand_fail:
-//  fprintf(stderr, "dn_expand fails\n");  
+/*  fprintf(stderr, "dn_expand fails\n"); */ 
   return -1;
 }
 
@@ -680,12 +809,88 @@ expand_fail:
  *
  dn_comp
 
- NOT IMPLEMENTED
+ Return -1 in case of overflow, but still fill buffer correctly.
+ We do not check the alphabet of the host names
+ nor the length of the compressed name and we
+ preserve the letter cases.
+
  *****************************************************************/
-int dn_comp(const char * a, u_char * b, int c, u_char ** d, u_char ** e)
+int dn_comp(const char * exp_dn, u_char * comp_dn, int length, 
+	    u_char ** dnptrs, u_char ** lastdnptr)
 {
-  errno = ENOSYS;
-  return -1;
+  u_char *cptr = comp_dn, *dptr, *lptr, *rptr;
+  unsigned int i, len;
+  u_char * const eptr = comp_dn + length - 1; /* Last valid */
+  
+  errno = EINVAL;
+
+  if (*exp_dn == '.' && !*(exp_dn + 1))
+    exp_dn++;
+  while (1) {
+    if (*exp_dn == '.' || cptr > eptr)
+      return -1;
+    if (*exp_dn == 0) {
+      *cptr++ = 0;
+      break;
+    }
+    /* Try to compress */
+    if (dnptrs) {
+      for (i = 1; dnptrs[i]; i++) {
+	dptr = dnptrs[i];
+	if (dptr >= comp_dn) /* Handle name.name */
+	  continue;
+	rptr = (u_char *) exp_dn;
+	len = *dptr++;
+	while (1) {
+	  do {
+	    if (*dptr++ != *rptr++)
+	      goto next_dn;
+	  } while (--len);
+	  len = *dptr++; 
+	  if (len == 0) { /* last label */
+	    if (!*rptr || (*rptr == '.' && !*(rptr + 1))) { /* Full match */
+	      len = (dnptrs[i] - dnptrs[0]) | 0xC000;
+	      /* Write pointer */
+	      *cptr++ = len >> 8;
+	      if (cptr > eptr)
+		return -1;
+	      *cptr++ = len;
+	      goto done;
+	    }
+	    goto next_dn;
+	  }
+	  if (*rptr++ != '.')
+	    goto next_dn;
+	  if (len >= 128 + 64) {
+	    dptr = dnptrs[0] + ((len - 128 - 64) << 8) + *dptr;
+	    len = *dptr++;
+	  }
+	}
+      next_dn: ;
+      }
+      /* Record label if asked and if space is available and if not too far off */
+      if (lastdnptr && (lastdnptr != &dnptrs[i]) && (cptr - dnptrs[0]) < 0xC000) {
+	dnptrs[i] = cptr;
+	dnptrs[i+1] = NULL;
+      }
+    }
+    /* Write label */
+    lptr = cptr++; /* Length byte */
+    rptr = (u_char *) exp_dn;
+    do {
+      if (cptr <= eptr)
+	*cptr++ = *rptr;
+    } while ((*++rptr != '.') && (*rptr != 0));
+    len = rptr - (u_char *) exp_dn;
+    if (len > MAXLABEL)
+      return -1;
+    *lptr = len;
+    exp_dn = (char *) rptr;
+    if (*exp_dn != 0)
+      exp_dn++; /* Skip over . */
+  }
+ done:
+  return cptr - comp_dn;
 }
 
 /*****************************************************************
